package config

import (
	"errors"
	"fmt"

	ma "gitlab.dms3.io/mf/go-multiaddr"
	peer "gitlab.dms3.io/p2p/go-p2p-core/peer"
)

// DefaultBootstrapAddresses are the hardcoded bootstrap addresses
// for DMS3. they are nodes run by the DMS3 team. docs on these later.
// As with all p2p networks, bootstrap is an important security concern.
//
// NOTE: This is here -- and not inside cmd/dms3/init.go -- because of an
// import dependency issue. TODO: move this into a config/default/ package.
var DefaultBootstrapAddresses = []string{
	// TODO: change these addresses for dms3:
	//      use dms3 domain DNS entries
	//      change mount mount point to dms3
	//      change node address hash
	//
	// FIXME TODO: replace with dms peer.IDs
	"/ip4/96.237.195.51/tcp/4101/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m", // moe.i.dms3.io
	"/ip4/96.237.195.52/tcp/4101/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m", // larry.i.dms3.io
	"/ip4/96.237.195.53/tcp/4101/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m", // curly.i.dms3.io
	"/ip4/96.237.195.54/tcp/4101/dms3/Qmeq45rCLjFt573aFKgLrcAmAMSmYy9WXTuetDsELM2r8m", // shemp.i.dms3.io

	/*
	   "/dnsaddr/bootstrap.dms3.io/p2p/QmNnooDu7bfjPFoTZYxMNLWUQJyrVwtbZg5gBMjTezGAJN",
	   "/dnsaddr/bootstrap.dms3.io/p2p/QmQCU2EcMqAqQPR2i9bChDtGNJchTbq5TbXJJ16u19uLTa",
	   "/dnsaddr/bootstrap.dms3.io/p2p/QmbLHAnMoJPWSCR5Zhtx6BHJX9KiKNN6tpvbUcqanj75Nb",
	   "/dnsaddr/bootstrap.dms3.io/p2p/QmcZf59bWwK5XFi76CZX8cbJ4BhTzzA3gU1ZjYZcYW3dwt",
	   "/ip4/104.131.131.82/tcp/4001/p2p/QmaCpDMGvV2BGHeYERUEnRQAwe3N8SzbUtfsmvsqQLuvuJ",      // mars.i.ip fs.io
	   "/ip4/104.131.131.82/udp/4001/quic/p2p/QmaCpDMGvV2BGHeYERUEnRQAwe3N8SzbUtfsmvsqQLuvuJ", // mars.i.ip fs.io
	*/
}

// ErrInvalidPeerAddr signals an address is not a valid peer address.
var ErrInvalidPeerAddr = errors.New("invalid peer address")

func (c *Config) BootstrapPeers() ([]peer.AddrInfo, error) {
	return ParseBootstrapPeers(c.Bootstrap)
}

// DefaultBootstrapPeers returns the (parsed) set of default bootstrap peers.
// if it fails, it returns a meaningful error for the user.
// This is here (and not inside cmd/dms3/init) because of module dependency problems.
func DefaultBootstrapPeers() ([]peer.AddrInfo, error) {
	ps, err := ParseBootstrapPeers(DefaultBootstrapAddresses)
	if err != nil {
		return nil, fmt.Errorf(`failed to parse hardcoded bootstrap peers: %s
This is a problem with the dms3 codebase. Please report it to the dev team.`, err)
	}
	return ps, nil
}

func (c *Config) SetBootstrapPeers(bps []peer.AddrInfo) {
	c.Bootstrap = BootstrapPeerStrings(bps)
}

// ParseBootstrapPeer parses a bootstrap list into a list of AddrInfos.
func ParseBootstrapPeers(addrs []string) ([]peer.AddrInfo, error) {
	maddrs := make([]ma.Multiaddr, len(addrs))
	for i, addr := range addrs {
		var err error
		maddrs[i], err = ma.NewMultiaddr(addr)
		if err != nil {
			return nil, err
		}
	}
	return peer.AddrInfosFromP2pAddrs(maddrs...)
}

// BootstrapPeerStrings formats a list of AddrInfos as a bootstrap peer list
// suitable for serialization.
func BootstrapPeerStrings(bps []peer.AddrInfo) []string {
	bpss := make([]string, 0, len(bps))
	for _, pi := range bps {
		addrs, err := peer.AddrInfoToP2pAddrs(&pi)
		if err != nil {
			// programmer error.
			panic(err)
		}
		for _, addr := range addrs {
			bpss = append(bpss, addr.String())
		}
	}
	return bpss
}
